/* 
 * SI - Sound Interface
 * Copyright (C) 2002 Matej Knopp
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __gem_list_h__
#define __gem_list_h__

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Double linked list
 */

typedef int  (* GemCompareFunc) (const void * ptr1, const void * ptr2);
typedef void (* GemDestroyFunc) (void * ptr);

typedef struct _GemList GemList;

struct _GemList
{
	GemList * prev;
	GemList * next;
	void    * data;
};

GemList  * gem_list_alloc     (void);
void       gem_list_free_node (GemList * list);

void       gem_list_free      (GemList * list);

GemList  * gem_list_prepend   (GemList * list, void * data);
GemList  * gem_list_append    (GemList * list, void * data);
GemList  * gem_list_insert    (GemList * list, int pos, void * data);

GemList  * gem_list_insert_after  (GemList * list, GemList * node, 
                                   void * data);
GemList  * gem_list_insert_before (GemList * list, GemList * node,
                                   void * data);

GemList  * gem_list_insert_sorted (GemList * list, void * data, 
								   GemCompareFunc comp);

GemList  * gem_list_node_at   (GemList * list, int pos);
void     * gem_list_data_at   (GemList * list, int pos);

GemList  * gem_list_remove_node (GemList * list, GemList * node);
GemList  * gem_list_delete_node (GemList * list, GemList * node);

GemList  * gem_list_remove     (GemList * list, const void * data);
GemList  * gem_list_remove_all (GemList * list, const void * data);

GemList  * gem_list_concat (GemList * list1, GemList * list2);

GemList  * gem_list_copy    (GemList * list);
GemList  * gem_list_reverse (GemList * list);

GemList  * gem_list_first  (GemList * list);
GemList  * gem_list_last   (GemList * list);
unsigned   gem_list_length (GemList * list);

void       gem_list_foreach (GemList * list, void (*func) (void * data));

GemList  * gem_list_find        (GemList * list, const void * data);
GemList  * gem_list_find_custom (GemList * list, const void * data, 
                                 GemCompareFunc comp);

GemList  * gem_list_sort (GemList * list, GemCompareFunc comp);

#define gem_list_next(list) (list ? list->next : NULL)
#define gem_list_prev(list) (list ? list->prev : NULL)

#ifdef __cplusplus
} /* extern "C" { */
#endif

#endif /* __gem_list_h__ */
